import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Scanner;

import static java.lang.System.out;

/**
   Command line interface for iTunes music store.
   
   @author Tom Austin
 */
public class CommandLineInterface
{
   //Fake storefront
   Facade facade;
   
   //Directory where music is downloaded
   private File musicDir;
   private File previewDir;
   
   //After a song has been listed, its information will be stored
   // in this map so that it can be downloaded.
   private Map<String,Song> songsListed;
   
   //Results from last performed search.
   private String[] lastSearchResults;
   
   //tracks whether a user is logged in or not
   //private boolean isLoggedIn;
   
   /**
      Default constructor.  Creates a directory in the user's home
      directory to stor the music downloads and previews
    */
   public CommandLineInterface()
   {
      this(System.getProperty("user.home") + "/iHope_iWont_bSued");
   }
   
   /**
      Constructor specifying the music directory
      @param musicDirName
    */
   public CommandLineInterface(String musicDirName)
   {
      //isLoggedIn = false;
      facade = new Facade(musicDirName);
      songsListed = new HashMap<String,Song>();
      lastSearchResults = null;
      
      musicDir = new File(musicDirName);
      if (!musicDir.exists()) musicDir.mkdir();
      
      previewDir = new File(musicDirName + "/_preview");
      previewDir.mkdir();
   }
   
   /**
      Runs the command line interface.
    */
   public void run()
   {
      Scanner scanner = new Scanner(System.in);
      scanner.useDelimiter(System.getProperty("line.separator"));
      
      String cmd = "";
      
      out.println();
      out.println();
      out.println("Welcome to ThaTune$!  Type 'exit' to exit the program, or 'help' for a list of commands.");
      out.println("Your songs and previews will be downloaded to " + musicDir.toString() + ".");
      out.print("thaTune$ ");
      
      //Keep reading commands until there is no input.
      // (This program may exit if the user types 'exit'.)
      while (scanner.hasNext())
      {
         String line = scanner.next();
         cmd = line.replaceFirst(" .*","");
         String args = line.replaceFirst(".*? ","");
         
         if (cmd.equalsIgnoreCase("exit") || cmd.equalsIgnoreCase("quit"))
         {
            System.out.println("  Thank you for using ThaTune$.  Good-bye.");
            out.println();
            out.println();
            System.exit(0);
         }
         else if (cmd.equalsIgnoreCase("help") || cmd.equalsIgnoreCase("usage"))
         {
            printUsage();
         }
         else if (cmd.equalsIgnoreCase("search"))
         {
            search(args);
         }
         else if (cmd.equalsIgnoreCase("preview"))
         {
            downloadPreview(args.split(" "));
         }
         else if (cmd.equalsIgnoreCase("previewAll"))
         {
             previewAll();
         }
         else if (cmd.equalsIgnoreCase("login"))
         {
             login(args);
         }
         else
         {
            out.println(cmd + " is not a recognized command.");
            printUsage();
         }
         
         //Print out the command prompt.
         out.print("thaTune$ ");
      }
   }
   
   /**
      @param args
    */
   private void login(String argsAsString)
   {
      String[] args = argsAsString.split(" ");
      if (args.length != 2)
      {
         out.println("  You need to specify a username and password to login.");
         return;
      }
      
      String username = args[0];
      String password = args[1];
      
      //We assume the account is an apple account, and not an aol account.
      try
      {
         facade.login(username, password, Facade.APPLE_ACCOUNT);
         out.println("  You are now logged in as " + username + ".");
      }
      catch(Exception e)
      {
         out.println("  Sorry, but the login did not succeed");
      }
      
      
   }

   /**
      Downloads previews for all songs from your last search.
    */
   private void previewAll()
   {
      if (lastSearchResults == null)
         out.println("  Sorry, there were no results from your last search to preview.");
      else downloadPreview(lastSearchResults);
   }
   
   /**
      Download previews for the specified songs
    */
   private void downloadPreview(String... songIds)
   {
      //If this is OS X, we'll build a command to open the files.
      String[] openCmd = new String[songIds.length + 1];
      int i=0;
      openCmd[i++] = "open";
      
      for (String id : songIds)
      {
         if (songsListed.containsKey(id))
         {
            Song s = songsListed.get(id);
            String previewLoc = s.downloadPreview(previewDir.getAbsolutePath());
            out.println("  Song " + id + " downloaded to " + previewLoc + ".");
            
            openCmd[i++] = previewLoc;
         }
         else out.println(  "Sorry, unable to find song ID " + id + ".");
      }
      
      //Play the previews in iTunes, if this is OS X
      if (System.getProperty("os.name").contains("Mac OS X"))
      {
         ProcessBuilder pb = new ProcessBuilder(openCmd);
         try
         {
            pb.start();
         }
         catch (IOException e)
         {
            e.printStackTrace();
         }
      }
   }

   /**
      Search for songs matching the given criteria.
    */
   private void search(String args)
   {
      Song[] matchingSongs = facade.search(args);
      
      //We will store the song IDs for previewAll to use later.
      lastSearchResults = new String[matchingSongs.length];
      int i=0;
      
      //If we got no matches, print a message and return.
      if (matchingSongs.length < 1)
      {
         out.println("  Sorry, no songs matching your terms were found");
         
         //Set lastSearchResults to null, so that previewAll will
         // display a message that there is nothing to play
         lastSearchResults = null;
         
         return;
      }
      
      out.println("The following songs matched your criteria:");
      for (Song s : matchingSongs)
      {
         out.println(s.toString());
         
         //Store the ID of the song for previewAll to use later.
         lastSearchResults[i++] = s.getID();
         
         //if this song was not listed before, store it
         // for easy access later.
         if (!songsListed.containsKey(s.getID()))
            songsListed.put(s.getID(), s);
      }
   }

   /**
      Prints instructions.
    */
   private void printUsage()
   {
      out.println("Usage:");
      out.println("  exit -- exit program");
      out.println("  help -- display this message");
      out.println("  login <username> <password> -- login to iTunes");
      out.println("  preview <songId list> -- download previews of songs.");
      out.println("  previewAll -- download previews for all songs from your last search.");
      out.println("  search <terms> -- search for songs from iTunes store.");
   }

   /**
      Runs the program.
    */
   public static void main(String[] args)
   {
      CommandLineInterface cli = new CommandLineInterface();
      cli.run();
   }
}
