import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLConnection;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Random;
import java.util.TimeZone;
import java.util.zip.GZIPInputStream;

import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
   This program serves as a fake iTunes client, allowing a user to interact with
   the iTunes server.
   
   This is based on the work of the Hyhm project and SharpMusique.
   
   @author Tom Austin
 */
public class Facade
{
   private String musicDir;
   //Login details needed to make purchases
   private int dsid = -1;
   private String passwordToken;
   private String guid;
   //143441 is for the U.S.
   private String storeFront = "143441";
   private String creditBal;
   private String creditDisp;
   private String freeSongBal;
   
   //identifies the type of account.
   public static final int APPLE_ACCOUNT = 0;
   public static final int AOL_ACCOUNT = 1;

   //Taken from http://phobos.apple.com/storeBag.xml.gz
   private static String SEARCH_URL = "http://ax.phobos.apple.com.edgesuite.net/WebObjects/MZSearch.woa/wa/DirectAction/search";
   
   //Taken from http://phobos.apple.com/secureBag.xml
   private static String AUTH_URL = "https://phobos.apple.com/WebObjects/MZFinance.woa/wa/authenticate";
   private static String BUY_URL = "https://phobos.apple.com/WebObjects/MZFinance.woa/wa/buyProduct";
   
   private static String USER_AGENT = "iTunes/5.0.1 (Macintosh; U; PPC Mac OS X 10.4.4)";
   //private static String USER_AGENT = "iTunes/6.0.4 (Macintosh; N; PPC Mac OS X 10.4.4)";
   
   /**
      Constructor.
      
      @param musicDir  Directory where ThaTunes can write files.
    */
   public Facade(String musicDir)
   {
      this.musicDir = musicDir;
      //Read GUID, or create it if it does not exist.
      File guidFile = new File(musicDir, ".guid");
      
      //Create guidFile if it does not exist
      if (!guidFile.exists())
      {
         guid = generateRandomGUID();
         writeFile(guidFile.getAbsolutePath(), guid.getBytes());
      }
      //If we have a guid already, we reuse it.
      else {
         try
         {
            FileInputStream in = new FileInputStream(guidFile);
            byte[] buf = new byte[(int) Math.min(guidFile.length(), 62)];
            in.read(buf);
            in.close();
      
            guid = new String(buf, 0, buf.length);
         }
         catch(IOException e)
         {
            e.printStackTrace();
         }
      }
   }
   
   /**
      Helper method to generate a random guid.
    */
   private static String generateRandomGUID()
   {
     long seed = System.currentTimeMillis();
     Random random = new Random(seed);
     StringBuffer id = new StringBuffer();

     for (int i = 0; i < 7; ++i) {
        if (i > 0) id.append(".");
        id.append(ThaUtil.toHexString(random.nextInt(), 8));
     }

     return id.toString();
   }
   
   /**
     Utility to write out to a file.
    */
   private static void writeFile(String fileName, byte[] data)
   {
      try
      {
         FileOutputStream out = new FileOutputStream(fileName);
         out.write(data);
         out.flush();
         out.close();
      }
      catch (IOException e)
      {
         e.printStackTrace();
      }
   }

   /**
      Allows a user to search for a given term from iTunes
      
      @param term  term to search for.
      @return  results.
    */
   public Song[] search(String term)
   {
      term = escapeURL(term);
      String results = request(SEARCH_URL, "?term=" + term);
      
      //Debug statements
      //System.out.println("***");
      //System.out.println(results);
      //System.out.println("***");
      
      return parseResults(results);
   }
   
   /**
      Parses XML search results and returns an array of the
      matching songs.
      
      @param results
      @return
    */
   private Song[] parseResults(String results)
   {
      //We will return an empty result list if there are any problems
      Song[] songMatches = new Song[0];
      
      DOMParser domParser = new DOMParser();

      StringReader stringReader = new StringReader(results);
      InputSource inputSource = new InputSource(stringReader);

      try
      {
         domParser.parse(inputSource);
         Document resultsDoc = domParser.getDocument();
         
         NodeList trackList = resultsDoc.getElementsByTagName("array").item(0).getChildNodes();
         
         //Half the nodes (about) are just junk...  They will be tossed.
         songMatches = new Song[trackList.getLength()/2];
         
         for (int i=0; i<trackList.getLength(); i++) {
            Node songNode = trackList.item(i);
            
            //Skip this node if it is junk.
            if (songNode.getNodeType() != Node.ELEMENT_NODE) continue;
            
            Song song = new Song();
            NodeList detailNodes = songNode.getChildNodes();
            String key=null;
            String val = null;
            
            for (int j=0; j<detailNodes.getLength(); j++) {
               Node n = detailNodes.item(j);
               if (n.getNodeName().equals("key")) {
                  key = n.getChildNodes().item(0).getNodeValue();
               }
               else if (n.getNodeName().equals("string")
                     || n.getNodeName().equals("integer")) {
                  val = n.getChildNodes().item(0).getNodeValue();
                  song.put(key,val);
                  key = val = "";
               }
            }
            //Every other node is crap, thanks to the oh so wonderful DOM API,
            // so i/2 gives us the right value.
            songMatches[i/2]=song;
         }
      }
      catch (SAXException e)
      {
         e.printStackTrace();
      }
      catch (IOException e)
      {
         e.printStackTrace();
      }
      
      return songMatches;
   }

   /**
      Convert string to the format needed in a url.
      
      @param s  string to be converted
      @return  a string in the format needed for a url.
    */
   private String escapeURL(String s)
   {
      StringBuilder sb = new StringBuilder("");
      // Go through character by character, escaping special characters as
      // needed.
      for (int i = 0; i < s.length(); i++)
      {
         char ch = s.charAt(i);
         if (ch == '&')
         {
            sb.append("&#38;");
         }
         else if (ch <= ' ' || ch == '#' || ch == '%' || ch == ';' || ch == '[' || ch == ']'
            || ch == '^' || ch == '`' || ch == '{' || ch == '}' || ch >= 0x7F)
         {
            sb.append(String.format("%%%2x", new Integer(ch)));
         }
         else
         {
            sb.append(ch);
         }
      }

      return sb.toString();
   }
   
   /**
      For methods that are not doing anything complex,
      getting the response as a String is more handy.
    */
   private String request(String url, String params) throws RuntimeException
   {
      byte[] content = request(url, params, null);
      
      if (content != null) try
      {
         return new String(content, "UTF-8");
      }
      catch (UnsupportedEncodingException e)
      {
         e.printStackTrace();
      }
      
      return null;
   }
   
   /**
      HTTP request -- This is nearly identical to the method
      for the JHymn project.  A 'cookie' parameter
      has been added to facilitate downloading encrypted
      purchased music.  Also, some of the logic has been
      cleaned up.
    */
   private byte[] request(String url, String params, String cookie) throws RuntimeException
   {
      byte[] content = null;
      URLConnection conn = null;
      
      TimeZone tz = TimeZone.getDefault();
      int offsetSeconds = tz.getOffset(System.currentTimeMillis()) / 1000;

      try
      {
         if (params == null) params = "";
         URL query = new URL(url + params);

         conn = query.openConnection();
         conn.setRequestProperty("Accept-Language", "en-us, en;q=0.50");
         conn.setRequestProperty("X-Apple-Tz", "" + offsetSeconds);

         if (passwordToken != null)
            conn.setRequestProperty("X-Token", passwordToken);
         
         if (cookie != null)
            conn.setRequestProperty("Cookie", cookie);

         conn.setRequestProperty("User-agent", USER_AGENT);

         if (dsid >= 0)
            conn.setRequestProperty("X-Dsid", "" + dsid);

         conn.setRequestProperty("X-Apple-Validation", computeValidator(url));
         conn.setRequestProperty("Accept-Encoding", "gzip, x-aes-cbc");

         if (storeFront != null)
            conn.setRequestProperty("X-Apple-Store-Front", storeFront);

         conn.connect();

         BufferedInputStream in = new BufferedInputStream(conn.getInputStream());
         int contentLength = conn.getContentLength();
         String contentEncoding = conn.getContentEncoding();
         int offset = 0;
         int bytesRead = 0;

         content = new byte[contentLength];

         while (offset < contentLength && bytesRead >= 0)
         {
            bytesRead = in.read(content, offset, contentLength - offset);
            if (bytesRead > 0) offset += bytesRead;
         }

         in.close();

         if (contentEncoding != null && contentEncoding.indexOf("x-aes-cbc") >= 0)
         {
            String ivString = conn.getHeaderField("x-apple-crypto-iv");
            byte[] iv = ThaUtil.hexStringToBytes(ivString);
            byte[] key;
            String twofish = conn.getHeaderField("x-apple-twofish-key");

            //The key may be scrambled for critical portions of the data
            if (twofish != null)
            {
               key = ThaUtil.hexStringToBytes(twofish);
               Crypto.descrambleForITMS(key);
            }
            else
            {
               int protocol = ThaUtil.to_int(conn.getHeaderField("x-apple-protocol-key"));

               if (protocol == 2) key = ThaUtil.fromBase64String("ip2tOZ+wFMExvmEYINeIlQ==");
               else if (protocol == 3) key = ThaUtil.fromBase64String("mNHiLKoNir1l0UOtJ1pe5w==");
               else throw new RuntimeException("Bad protocol " + protocol
                     + " in response from " + url);
            }

            if (iv == null || !Crypto.aesDecrypt(content, key, iv)) {
               throw new RuntimeException("Error in decrypting response from " + url);
            }
         }

         if (contentEncoding != null && contentEncoding.indexOf("gzip") >= 0)
         {
            GZIPInputStream gzin = new GZIPInputStream(new ByteArrayInputStream(content));
            byte[] unzipped = new byte[contentLength];
            int blockSize = 256;

            offset = 0;

            do
            {
               if (unzipped.length - offset < blockSize) {
                  unzipped = ThaUtil.resizeByteArray(unzipped, unzipped.length * 3/2 + blockSize);
               }

               bytesRead = gzin.read(unzipped, offset, blockSize);

               if (bytesRead > 0) offset += bytesRead;
            } while (bytesRead > 0);

            gzin.close();
            content = ThaUtil.resizeByteArray(unzipped, offset);
         }
      }
      catch (IOException e)
      {
         String errorMsg = "Error in receiving response from " + url + ": ";
         String name;
         String value;
         boolean gotCode = false;

         for (int i = 0;; ++i)
         {
            name = conn.getHeaderFieldKey(i);
            value = conn.getHeaderField(i);

            if (name == null && value == null) break;
            else if (name == null)
            {
               errorMsg += value;
               gotCode = true;
               break;
            }
         }

         if (!gotCode || !url.equals(e.getMessage()))
         {
            if (gotCode) errorMsg += "\n";
            errorMsg += e.getMessage();
         }

         throw new RuntimeException(errorMsg);
      }

      return content;
   }
   
   /**
      Validates the computer, based on the version
      of iTunes we are pretending to be, the url,
      and a randomly generated hex value.
      
      @param url
      @return String made of a random hex value, a dash, and the
          md5 hash. 
    */
   private String computeValidator(String url)
   {
      long randomNum = (long) (Math.random() * 0x100000000L);
      String randomStr = ThaUtil.toHexString(randomNum, 8);
      byte[] abRandom = randomStr.getBytes();
      byte[] abStatic = ThaUtil.fromBase64String("ROkjAaKid4EUF5kGtTNn3Q==");
      int pos = 0;
      String validator;

      for (int i = 0; i < 3 && pos >= 0; ++i)
         pos = url.indexOf('/', pos + 1);

      byte[] abURL = url.substring(pos).getBytes();
      byte[] abUA = USER_AGENT.getBytes();
      byte[] hash = null;

      try
      {
         MessageDigest md = MessageDigest.getInstance("MD5");

         md.update(abURL);
         md.update(abUA);
         md.update(abStatic);

         hash = md.digest(abRandom);
      }
      catch (NoSuchAlgorithmException e)
      {
         e.printStackTrace();
      }

      validator = randomStr + "-" + ThaUtil.bytesToString(hash);

      return validator;
   }
   
   /**
      This logs in to the iTunes music store.
      @param userID  login name for the account.
      @param password  password for the account.
      @param accountKind  Apple or AOL (use class public constants).
    */
   public void login(String userID, String password, int accountKind)
   {
     String params = "?appleId=" + ThaUtil.urlEncode(userID) +
                     "&password=" + ThaUtil.urlEncode(password) +
                     "&accountKind=" + accountKind +
                     "&attempt=1" +
                     "&guid=" + ThaUtil.urlEncode(guid);

     String response = request(AUTH_URL, params);
     
     //debug statement
     //System.out.println(response);
     
     Document document = ThaUtil.getDocumentFromString(response);
     if (document == null) throw new RuntimeException("Bad 'authenticateAccount' response.");

     String docType = ThaUtil.getValueForKey(document, "jingleDocType");
     String custMsg = ThaUtil.getValueForKey(document, "customerMessage");

     passwordToken = ThaUtil.getValueForKey(document, "passwordToken");
     dsid = ThaUtil.to_int(ThaUtil.getValueForKey(document, "dsPersonId"), -1);
     
     creditBal = ThaUtil.getValueForKey(document, "creditBalance");
     creditDisp = ThaUtil.getValueForKey(document, "creditDisplay");
     freeSongBal = ThaUtil.getValueForKey(document, "freeSongBalance");

     if (!"authenticationSuccess".equalsIgnoreCase(docType))
       throw new RuntimeException("Login failed: " + custMsg);

     if (passwordToken == null)
       throw new RuntimeException("Login error: no passwordToken");

     if (dsid < 0)
       throw new RuntimeException("Login error: no dsPersonId");
   }
   
   /**
      Purchases a song, but does not download it.
      
      @param s  Song to buy.
    */
   public void buySong(Song s)
   {
      String params = "?" + s.get("buyParams") +
                      "&creditBalance=" + ThaUtil.urlEncode(creditBal) +
                      "&creditDisplay=" + ThaUtil.urlEncode(creditDisp) +
                      "&freeSongBalance=" + ThaUtil.urlEncode(freeSongBal) +
                      "&guid=" + ThaUtil.urlEncode(guid) +
                      "&rebuy=false&buyWithoutAuthorization=true" +
                      "&wasWarnedAboutFirstTimeBuy=true";
      
      String response = request(BUY_URL, params);
      
      //Writing out the critical information, in case there was a problem
      String outputFile = musicDir + "/" + s.getFileName() + "_receipt.xml";
      try
      {
         FileOutputStream out = new FileOutputStream(outputFile);
         out.write(response.getBytes());
         out.flush();
         out.close();
      }
      catch (IOException e)
      {
         e.printStackTrace();
      }
      
      //Debug statement
      //System.out.println(response);
      
      Document document = ThaUtil.getDocumentFromString(response);
      if (document == null) throw new RuntimeException("Bad 'buyProduct' response.");
      
      //Tells us whether the transaction was a success or not.
      String docType = ThaUtil.getValueForKey(document, "jingleDocType");
      //Error message, if there is one.
      String custMsg = ThaUtil.getValueForKey(document, "customerMessage");
      
      //Update balances
      creditBal = ThaUtil.getValueForKey(document, "creditBalance");
      creditDisp = ThaUtil.getValueForKey(document, "creditDisplay");
      freeSongBal = ThaUtil.getValueForKey(document, "freeSongBalance");
      
      if (!"purchaseSuccess".equalsIgnoreCase(docType))
         throw new RuntimeException("Purchase failed: " + custMsg);
      
      //Store the url and keys for the song
      String downloadURL = ThaUtil.getValueForKey(document, "URL");
      s.put("URL", downloadURL);
      String downloadKey = ThaUtil.getValueForKey(document, "downloadKey");
      s.put("downloadKey", downloadKey);
      String encryptionKey= ThaUtil.getValueForKey(document, "encryptionKey");
      s.put("encryptionKey", encryptionKey);
   }
   
   /**
      Downloads a purchased song.
      WARNING!!!  Currently this fails to generate a legitimate mp4 file.
    */
   public void downloadSong(Song s)
   {
      String downloadKey = s.get("downloadKey");
      byte[] data = request(s.get("URL"), null, "downloadKey=" + downloadKey);
      
      //Write out the raw, encrypted file in case of prablems
      String rawFileName = musicDir + "/" + s.getFileName() + ".ENCR";
      writeFile(rawFileName, data);
      
      //   Decrypt the data
      byte[] encrKey = ThaUtil.hexStringToBytes(s.get("encryptionKey"));
      //IV seems to be the first 16 bytes of data.
      byte[] iv = new byte[16];
      System.arraycopy(data, 0, iv, 0, iv.length);
      //This will decrypt the data
      Crypto.aesDecrypt(data, encrKey,iv);
      
      //Writing the information to mp4 format.
      byte[] mp4=null;
      try
      {
         byte[] ftyp = "\0\0\0\0ftypM4A \0\0\0\0".getBytes("US-ASCII");
         
         //append the data to the ftyp bytes
         mp4 = new byte[ftyp.length + data.length];
         System.arraycopy(ftyp, 0, mp4, 0, ftyp.length);
         System.arraycopy(data, 0, mp4, ftyp.length, data.length);
         
         //Replace the 'moov' atom
         byte[] moov = "moov".getBytes("US-ASCII");
         out: for(int i=0; i<mp4.length-4; i++)
         {
            if( mp4[i+0]==moov[0] && mp4[i+1]==moov[1] && mp4[i+2]==moov[2] && mp4[i+3]==moov[3])
            {
               //No unsigned int in Java, so making do...
               long val = i-4;
               byte[] offb = new byte[4];
               for (i=3; i>=0; i--) {
                  offb[i] = (byte) (val & 0xff);
                  val >>= 8;
               }

               System.arraycopy(offb, 0, mp4, 0, offb.length);
               break out;
            }
         }
         String outputFile = musicDir + "/" + s.getFileName();
         writeFile(outputFile, mp4);
      }
      catch (UnsupportedEncodingException e)
      {
         e.printStackTrace();
      }
   }

   /**
      Test main method.  Used only for test and demo purposes.
      It searches for a song, buys it, and downloads it.
      Currently, the login name and password are invalid, so it
      will just fail.
    */
   public static void main(String[] args)
   {
      Facade fac = new Facade("./iHope_iWont_bSued");
      
      Song[] songsFound = fac.search("weird al a complicated song");
      
      Song buyMe = null; 
      
      System.out.println("Songs matching your query:");
      for (Song s : songsFound)
      {
         System.out.println(s.toString());
         if (s.getID().equals("2322101"))
         {
            buyMe = s;
         }
      }
      
      //The first 2 parameters should be replaced with username and password.
      fac.login("*", "*", APPLE_ACCOUNT);
      fac.buySong(buyMe);
      fac.downloadSong(buyMe);
   }

}
